/*
 * Copyright (c) 2014 Cloudera, Inc. All rights reserved.
 */

package com.cloudera.keytrustee;

import com.cloudera.keytrustee.crypto.Fingerprint;
import com.cloudera.keytrustee.impl.ClientFactoryImpl;
import org.json.JSONArray;

import java.util.Collection;
import java.util.concurrent.TimeoutException;

import java.lang.reflect.Proxy;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;

/**
 * Dynamic proxy for KeyTrustee ClientConnection interface with specific support for
 * TrusteeClientConnectionPool. Handles return of ClientConnection objects to the pool after
 * invocation in successful, exceptional and error conditions to reduce the amount of boilerplate client code.
 */
public class ClientConnectionInvocationHandler implements InvocationHandler {

  private final ClientConnection keyTrusteeConnection;
  private final TrusteeClientConnectionPool keyTrusteeConnectionPool;

  private ClientConnectionInvocationHandler(ClientConnection keyTrusteeConnection) {
    this.keyTrusteeConnectionPool = null;
    this.keyTrusteeConnection = keyTrusteeConnection;
  }

  private ClientConnectionInvocationHandler(TrusteeClientConnectionPool keyTrusteeConnectionPool) {
    this.keyTrusteeConnectionPool = keyTrusteeConnectionPool;
    this.keyTrusteeConnection = null;

  }

  public Object invoke(Object proxy, Method method, Object[] args)
          throws Throwable {
    boolean connectionReleased = true;
    ClientConnection keyTrusteeConnection = null;
    try {
      keyTrusteeConnection = getConnection();
      connectionReleased = false;
      Object retVal = method.invoke(keyTrusteeConnection, args);
      connectionReleased = releaseConnection(keyTrusteeConnection);
      return retVal;
    } catch (java.lang.reflect.InvocationTargetException e) {
      throw e.getCause();
    } finally {
        if (!connectionReleased) releaseConnection(keyTrusteeConnection);
    }
  }

  public static ClientConnection createClientConnection(ClientConnection keyTrusteeConnection) {
    return (ClientConnection)(Proxy.newProxyInstance(ClientConnection.class.getClassLoader(),
            new Class[] {ClientConnection.class},
            new ClientConnectionInvocationHandler(keyTrusteeConnection)));
  }

  public static ClientConnection createClientConnection(TrusteeClientConnectionPool keyTrusteeConnectionPool) {
    return (ClientConnection)(Proxy.newProxyInstance(ClientConnection.class.getClassLoader(),
            new Class[] {ClientConnection.class},
            new ClientConnectionInvocationHandler(keyTrusteeConnectionPool)));
  }

  public int getIdleConnectionCount() {
    if (null != keyTrusteeConnectionPool) {
      return keyTrusteeConnectionPool.getNumIdleConnections();
    } else {
      return -1;
    }
  }

  public int getActiveConnectionCount() {
    if (null != keyTrusteeConnectionPool) {
      return keyTrusteeConnectionPool.getNumActiveConnections();
    } else {
      return -1;
    }
  }

  private ClientConnection getConnection() {
    if (null != this.keyTrusteeConnectionPool) {
      return this.keyTrusteeConnectionPool.getConnection();
    } else {
      return this.keyTrusteeConnection;
    }
  }

  private boolean releaseConnection(ClientConnection clientConnection) {
    if (null != this.keyTrusteeConnectionPool) {
      return this.keyTrusteeConnectionPool.releaseConnection(clientConnection);
    } else {
      return true; //do nothing
    }
  }
}
