package org.apache.hadoop.crypto.key.kms.server;

import com.cloudera.keytrustee.TrusteeKeyProvider;
import com.google.common.base.Preconditions;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.crypto.key.KeyProvider;
import org.apache.hadoop.crypto.key.KeyProviderCryptoExtension;
import org.apache.hadoop.crypto.key.KeyProviderFactory;
import org.apache.hadoop.crypto.key.kms.KMSRESTConstants;
import org.apache.hadoop.security.AccessControlException;
import org.apache.hadoop.security.UserGroupInformation;
import org.apache.hadoop.security.authorize.AuthorizationException;
import org.apache.hadoop.security.token.delegation.web.HttpUserGroupInformation;

import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.QueryParam;
import javax.ws.rs.core.Response;
import java.net.URI;
import java.net.URISyntaxException;
import java.security.PrivilegedExceptionAction;
import java.util.List;

/**
 * Class providing the REST bindings, via Jersey, for the Trustee specific 
 * operations (unDelete and purge). Based on org.apache.hadoop.crypto.key.kms.server.KMS.
 */
@Path(TrusteeRestApi.TRUSTEE_RESOURCE)
public class TrusteeRestApi {

  private static final Log LOG = LogFactory.getLog(TrusteeRestApi.class);

  public static final String TRUSTEE_RESOURCE = 
      KMSRESTConstants.SERVICE_VERSION + "/trustee";

  public static final String TRUSTEE_OP = "trusteeOp";

  public static final String UNDELETE = "undelete";

  public static final String PURGE = "purge";

  public static final String MIGRATE = "migrate";

  private static final String UNAUTHORIZED_MSG_WITH_KEY =
          "User:%s not allowed to do '%s' on '%s'";

  private static final String UNAUTHORIZED_MSG_WITHOUT_KEY =
          "User:%s not allowed to do '%s'";

  // this should be a TrusteeKeyProvider var
  private KeyProvider provider;
  private boolean validTrusteeProvider = false;
  private TrusteeACLs acls;

  public TrusteeRestApi() throws Exception {
    Configuration kmsConf = KMSConfiguration.getKMSConf();
    // List<KeyProvider> providers = KeyProviderFactory.getProviders(kmsConf);

    // intializing the KeyProvider
    String providerString = kmsConf.get(KMSConfiguration.KEY_PROVIDER_URI);

    if (providerString == null) {
      throw new IllegalStateException("No KeyProvider has been defined");
    }
    provider =
            KeyProviderFactory.get(new URI(providerString), kmsConf);
    /*
    if (providers.isEmpty()) {
      throw new IllegalStateException("No KeyProvider has been defined");
    }
    if (providers.size() > 1) {
      LOG.warn("There is more than one KeyProvider configured '{" + kmsConf.get(KeyProviderFactory.KEY_PROVIDER_PATH) +
                      "}', using the first provider");
    }
    provider = providers.get(0);
    */
    
    if (provider instanceof TrusteeKeyProvider) {
      validTrusteeProvider = true;
    } else if (provider instanceof KeyProviderCryptoExtension) {
      provider = ((KeyProviderCryptoExtension)provider);
    } else {
      if (LOG.isErrorEnabled()) {
        LOG.error("Trustee REST API will not be active as provider is not an instance of TrusteeKeyProvider.");
      }
    }

    acls = new TrusteeACLs();
  }

  private void assertAccess(TrusteeACLs.Type aclType, UserGroupInformation ugi, String operation,
      String key) throws AccessControlException {
    // if unauthorized log in kmsaudit and throw AccessControlException
    if (!acls.hasAccess(aclType, ugi)) {
      KMSWebApp.getUnauthorizedCallsMeter().mark();
      //TODO: Next line will not compile until change to KMSAudit.unauthorized
      // KMSWebApp.getKMSAudit().unauthorized(ugi, operation, key);
      throw new AuthorizationException(String.format(
              (key != null) ? UNAUTHORIZED_MSG_WITH_KEY
                      : UNAUTHORIZED_MSG_WITHOUT_KEY,
              ugi.getShortUserName(), operation, key));
    }

  }

  @POST
  @Path(KMSRESTConstants.KEY_RESOURCE + "/{name:.*}")
  public Response keyTrusteeOperationOnKey(@PathParam("name") final String name,
      @QueryParam(TRUSTEE_OP) final String trusteeOp)
      throws Exception {
    KMSWebApp.getAdminCallsMeter().mark();
    UserGroupInformation user = HttpUserGroupInformation.get();

    //Preconditions.checkNotNull(name, "name");
    Preconditions.checkNotNull(trusteeOp, TRUSTEE_OP);
    Preconditions.checkArgument(trusteeOp.equals(UNDELETE) ||
            trusteeOp.equals(PURGE) || trusteeOp.equals(MIGRATE), TRUSTEE_OP + " must be '" + UNDELETE +
            "' or '" + PURGE + "'" + "' or '" + MIGRATE + "'");

    if (trusteeOp.equals(UNDELETE)) {
      assertAccess(TrusteeACLs.Type.UNDELETE, user, trusteeOp, name);
    } else if (trusteeOp.equals(PURGE)) {
      assertAccess(TrusteeACLs.Type.PURGE, user, trusteeOp, name);
    } else if (trusteeOp.equals(MIGRATE)) {
      assertAccess(TrusteeACLs.Type.MIGRATE, user, trusteeOp, name);
    }

    user.doAs(new PrivilegedExceptionAction<Void>() {
      @Override
      public Void run() throws Exception {
        if (validTrusteeProvider) {
          TrusteeKeyProvider tkp = (TrusteeKeyProvider) provider;
          if (trusteeOp.equals(UNDELETE)) {
            tkp.unDeleteKey(name);
          } else if (trusteeOp.equals(PURGE)) {
            tkp.purgeKey(name);
          }  else if (trusteeOp.equals(MIGRATE)) {
            initiateMigration(tkp);
          } else {
            throw new IllegalArgumentException("It cannot happen, " + TRUSTEE_OP +
                    ": " + trusteeOp);
          }
          provider.flush();
          return null;
        } else {
          if (LOG.isErrorEnabled()) {
            LOG.error("Trustee REST API is not active as provider is not an instance of TrusteeKeyProvider.");
            // throw an exception back to the client?
          }
          return null;
        }
      }
    });

    // audit OK operation

    return Response.ok().build();
  }

  private void initiateMigration(TrusteeKeyProvider tkp) throws java.io.IOException {
    URI[] uris = tkp.getSpecifiedProviderURIs();
    Preconditions.checkNotNull(uris, "Migration URI array is null.");
    if (uris.length < 2) {
      throw new java.io.IOException("Migration URI array must contain at least two entries (TKP and one or more others).");
    } else {
      for (int i = 0 ; i < uris.length; i++) {
        URI uri = uris[i];
        LOG.info("Migration attempt on URI: " + uri);
        if (i == 0) {
          if (!(TrusteeKeyProvider.SCHEME_NAME.equals(uri.getScheme()))) {
            throw new java.io.IOException("First element in Migration URI array must be a keytrustee:// URI.");
          }
        } else {
            KeyProvider sourceProvider = KeyProviderFactory.get(uri, KMSConfiguration.getKMSConf());
            LOG.info("Migration attempt on provider: " + sourceProvider);
            tkp.migrateKeys(sourceProvider);
        }
      }
    }

  }

}
